export interface Resource<T, D extends NonNullable<unknown>> {
    resolve: Resolve<T, D>;
    finalize: (item: T) => Promise<void>;
}
export interface Definition {
    [key: string]: {
        dependencies: {
            [key: string]: any;
        };
        value: any;
    };
}
export type Injector<T extends {
    [key: string]: any;
}> = {
    [P in keyof T]: Promise<T[P]>;
};
export type Resolve<V, D extends NonNullable<unknown>> = (injector: Injector<D>) => Promise<V>;
export type PromisesHandler = (ps: Promise<void>[]) => Promise<void>;
type Resolvable<V, D extends NonNullable<unknown>> = (injector: Injector<D>) => V | Promise<V>;
type ExactOneValue<T> = {
    [P in keyof T]: Exclude<T[keyof T], T[P]> extends never ? T[P] : never;
}[keyof T];
type DependentValue<T extends Definition, K> = ExactOneValue<{
    [P in keyof T]: K extends keyof T[P]['dependencies'] ? T[P]['dependencies'][K] : never;
}>;
type BoundKeys<T extends Definition> = Extract<{
    [P in keyof T]: keyof T[P]['dependencies'];
}[keyof T], keyof T>;
type MissingKeys<T extends Definition> = Exclude<{
    [P in keyof T]: keyof T[P]['dependencies'];
}[keyof T], keyof T>;
type MissingDependencies<T extends Definition> = {
    [P in MissingKeys<T>]: DependentValue<T, P>;
};
type ConflictedKeys<T extends Definition> = {
    [P in BoundKeys<T>]: T[P]['value'] extends DependentValue<T, P> ? never : P;
}[BoundKeys<T>];
type Requirements<T extends Definition> = MissingDependencies<T> & {
    [P in ConflictedKeys<T>]: never;
};
export interface Result<T extends Definition> {
    container: Container<T>;
    finalize: (promisesHandler?: PromisesHandler) => Promise<void>;
}
export type Container<T extends Definition> = {
    [P in keyof T]: T[P]['value'];
};
export type Underlying<T extends Definition> = {
    [P in keyof T]: Resource<T[P]['value'], T[P]['dependencies']>;
};
/**
 * Design represents key-value styled dependency graph which can detect which dependent key is missing at compile time.
 *
 * i.g.)
 * ```
 * type HasNumber = { numberKey: number }
 *
 * // NOTE: `(injector: Injector<T>) => Promise<U>` represents dependencies for a value U.
 * // Injector<T> resolves value T[keyof T] by calling `injector.key()` asynchronously.
 * const detectStringFromNumber =
 *   async (injector: Injector<HasNumber>) => `number is ${await injector.numberKey}`;
 *
 * const design = Design.empty
 *   .bind('stringFromNumber', detectStringFromNumber);
 *
 * design.resolve({}) // compile error
 * design.resolve({ numberKey: 123 }) // returns Promise({ numberKey: 123, stringFromNumber: 'number is 123' })
 * ```
 *
 * see specs for more examples!
 *
 */
export declare class Design<T extends Definition> {
    readonly design: Underlying<T>;
    private constructor();
    bind: <K extends string, V, D extends {} = {}>(key: K, resolvable: Resolvable<V, Container<T> & D>, finalize?: (item: V) => Promise<void>) => Design<T & { [key in K]: {
        dependencies: D;
        value: V;
    }; }>;
    bindResource: <K extends string, V extends {
        finalize(): Promise<void>;
    }, D extends {} = {}>(key: K, resolvable: Resolvable<V, Container<T> & D>) => Design<T & { [key in K]: {
        dependencies: D;
        value: V;
    }; }>;
    merge: <U extends Definition>(that: Design<U>) => Design<T & U>;
    resolve: (requirements: Requirements<T>) => Promise<Result<T>>;
    use: (requirements: Requirements<T>) => <A>(f: (container: Container<T>) => Promise<A>) => Promise<A>;
    static pure: <U extends {
        [key: string]: any;
    }>(mapping: U) => Design<{ [P in keyof U]: {
        dependencies: NonNullable<unknown>;
        value: U[P];
    }; }>;
    static empty: Design<NonNullable<unknown>>;
    static bind: <K extends string, V, D extends {} = {}>(key: K, resolvable: Resolvable<V, Container<{}> & D>, finalize?: (item: V) => Promise<void>) => Design<{ [key in K]: {
        dependencies: D;
        value: V;
    }; }>;
    static bindResource: <K extends string, V extends {
        finalize(): Promise<void>;
    }, D extends {} = {}>(key: K, resolvable: Resolvable<V, Container<{}> & D>) => Design<{ [key in K]: {
        dependencies: D;
        value: V;
    }; }>;
}
export {};
